@@ -111,11 +111,30 @@ AVPixelFormat FrameWriter::get_input_format()
         AV_PIX_FMT_BGR0 : AV_PIX_FMT_RGB0;
 }
 
+AVPixelFormat FrameWriter::lookup_pixel_format(std::string pix_fmt)
+{
+    AVPixelFormat fmt = av_get_pix_fmt(pix_fmt.c_str());
+
+    if (fmt != AV_PIX_FMT_NONE)
+      return fmt;
+
+    std::cerr << "Failed to find the pixel format: " << pix_fmt << std::endl;
+    std::exit(-1);
+}
+
 AVPixelFormat FrameWriter::choose_sw_format(AVCodec *codec)
 {
-    /* First case: if the codec supports getting the appropriate RGB format
-     * directly, we want to use it since we don't have to convert data */
     auto in_fmt = get_input_format();
+
+    if (!params.pix_fmt.empty())
+        return lookup_pixel_format(params.pix_fmt);
+
+    /* For codecs such as rawvideo no supported formats are listed */
+    if (!codec->pix_fmts)
+        return in_fmt;
+
+    /* If the codec supports getting the appropriate RGB format
+     * directly, we want to use it since we don't have to convert data */
     if (is_fmt_supported(in_fmt, codec->pix_fmts))
         return in_fmt;
 
@@ -135,7 +154,7 @@ void FrameWriter::init_video_stream()
     AVCodec* codec = avcodec_find_encoder_by_name(params.codec.c_str());
     if (!codec)
     {
-        std::cerr << "Failed to find the given codec" << std::endl;
+        std::cerr << "Failed to find the given codec: " << params.codec << std::endl;
         std::exit(-1);
     }
 
@@ -151,30 +170,37 @@ void FrameWriter::init_video_stream()
     videoCodecCtx->height = params.height;
     videoCodecCtx->time_base = (AVRational){ 1, FPS };
 
+    if (params.bframes != -1)
+        videoCodecCtx->max_b_frames = params.bframes;
+
     if (params.codec.find("vaapi") != std::string::npos)
     {
         videoCodecCtx->pix_fmt = AV_PIX_FMT_VAAPI;
         init_hw_accel();
         videoCodecCtx->hw_frames_ctx = av_buffer_ref(hw_frame_context);
+
+        if (params.force_yuv)
+            init_sws(AV_PIX_FMT_YUV420P);
     } else
     {
         videoCodecCtx->pix_fmt = choose_sw_format(codec);
         std::cout << "Choosing pixel format " <<
             av_get_pix_fmt_name(videoCodecCtx->pix_fmt) << std::endl;
-        init_sws();
+        init_sws(videoCodecCtx->pix_fmt);
     }
 
     if (fmtCtx->oformat->flags & AVFMT_GLOBALHEADER)
         videoCodecCtx->flags |= AV_CODEC_FLAG_GLOBAL_HEADER;
 
-    int err;
-    if ((err = avcodec_open2(videoCodecCtx, codec, &options)) < 0)
+    int ret;
+    char err[256];
+    if ((ret = avcodec_open2(videoCodecCtx, codec, &options)) < 0)
     {
-        std::cerr << "avcodec_open2 failed: " << averr(err) << std::endl;
+        av_strerror(ret, err, 256);
+        std::cerr << "avcodec_open2 failed: " << err << std::endl;
         std::exit(-1);
     }
     av_dict_free(&options);
-    videoStream->time_base = (AVRational){ 1, FPS };
 }
 
 static uint64_t get_codec_channel_layout(AVCodec *codec)
@@ -244,7 +270,7 @@ void FrameWriter::init_audio_stream()
     swrCtx = swr_alloc();
     if (!swrCtx)
     {
-        std::cerr << "Faild to allocate swr context" << std::endl;
+        std::cerr << "Failed to allocate swr context" << std::endl;
         std::exit(-1);
     }
 
@@ -276,18 +302,22 @@ void FrameWriter::init_codecs()
     }
 
     AVDictionary *dummy = NULL;
-    if (avformat_write_header(fmtCtx, &dummy) != 0)
+    char err[256];
+    int ret;
+    if ((ret = avformat_write_header(fmtCtx, &dummy)) != 0)
     {
         std::cerr << "Failed to write file header" << std::endl;
+        av_strerror(ret, err, 256);
+        std::cerr << err << std::endl;
         std::exit(-1);
     }
     av_dict_free(&dummy);
 }
 
-void FrameWriter::init_sws()
+void FrameWriter::init_sws(AVPixelFormat format)
 {
     swsCtx = sws_getContext(params.width, params.height, get_input_format(),
-        params.width, params.height, videoCodecCtx->pix_fmt,
+        params.width, params.height, format,
         SWS_FAST_BILINEAR, NULL, NULL, NULL);
 
     if (!swsCtx)
@@ -297,32 +327,52 @@ void FrameWriter::init_sws()
     }
 }
 
+static const char* determine_output_format(const FrameWriterParams& params)
+{
+    if (!params.muxer.empty())
+        return params.muxer.c_str();
+
+    if (params.file.find("rtmp") == 0)
+        return "flv";
+
+    if (params.file.find("udp") == 0)
+        return "mpegts";
+
+    return NULL;
+}
+
 FrameWriter::FrameWriter(const FrameWriterParams& _params) :
     params(_params)
 {
     if (params.enable_ffmpeg_debug_output)
         av_log_set_level(AV_LOG_DEBUG);
 
+#ifdef HAVE_LIBAVDEVICE
+    avdevice_register_all();
+#endif
+
     // Preparing the data concerning the format and codec,
     // in order to write properly the header, frame data and end of file.
     this->outputFmt = av_guess_format(NULL, params.file.c_str(), NULL);
-    if (!outputFmt)
-    {
-        std::cerr << "Failed to guess output format for file " << params.file << std::endl;
-        std::exit(-1);
-    }
-
-    if (avformat_alloc_output_context2(&this->fmtCtx, NULL, NULL, params.file.c_str()) < 0)
+    auto streamFormat = determine_output_format(params);
+    auto context_ret = avformat_alloc_output_context2(&this->fmtCtx, NULL,
+        streamFormat, params.file.c_str());
+    if (context_ret < 0)
     {
         std::cerr << "Failed to allocate output context" << std::endl;
         std::exit(-1);
     }
 
+#ifndef HAVE_OPENCL
+    if (params.opencl)
+        std::cerr << "This version of wf-recorder was built without OpenCL support. Ignoring OpenCL option." << std::endl;
+#endif
+
     init_codecs();
 
     encoder_frame = av_frame_alloc();
     if (hw_device_context) {
-        encoder_frame->format = get_input_format();
+        encoder_frame->format = params.force_yuv ? AV_PIX_FMT_YUV420P : get_input_format();
     } else {
         encoder_frame->format = videoCodecCtx->pix_fmt;
     }
@@ -351,6 +401,61 @@ FrameWriter::FrameWriter(const FrameWriterParams& _params) :
     }
 }
 
+void FrameWriter::convert_pixels_to_yuv(const uint8_t *pixels,
+    const uint8_t *formatted_pixels, int stride[])
+{
+    bool y_invert = (pixels != formatted_pixels);
+    bool converted_with_opencl = false;
+
+#ifdef HAVE_OPENCL
+    if (params.opencl && params.force_yuv)
+    {
+        int r = opencl->do_frame(pixels, encoder_frame,
+            get_input_format(), y_invert);
+
+        converted_with_opencl = (r == 0);
+    }
+#else
+    /* Silence compiler warning when opencl is disabled */
+    (void)(y_invert);
+#endif
+
+    if (!converted_with_opencl)
+    {
+        sws_scale(swsCtx, &formatted_pixels, stride, 0, params.height,
+            encoder_frame->data, encoder_frame->linesize);
+    }
+}
+
+void FrameWriter::encode(AVCodecContext *enc_ctx, AVFrame *frame, AVPacket *pkt)
+{
+    int ret;
+
+    /* send the frame to the encoder */
+    ret = avcodec_send_frame(enc_ctx, frame);
+    if (ret < 0)
+    {
+        fprintf(stderr, "error sending a frame for encoding\n");
+        return;
+    }
+
+    while (ret >= 0)
+    {
+        ret = avcodec_receive_packet(enc_ctx, pkt);
+        if (ret == AVERROR(EAGAIN) || ret == AVERROR_EOF)
+        {
+            return;
+        }
+        if (ret < 0)
+        {
+            fprintf(stderr, "error during encoding\n");
+            return;
+        }
+
+        finish_frame(enc_ctx, *pkt);
+    }
+}
+
 void FrameWriter::add_frame(const uint8_t* pixels, int64_t usec, bool y_invert)
 {
     /* Calculate data after y-inversion */
@@ -367,8 +472,12 @@ void FrameWriter::add_frame(const uint8_t* pixels, int64_t usec, bool y_invert)
 
     if (hw_device_context)
     {
-        encoder_frame->data[0] = (uint8_t*)formatted_pixels;
-        encoder_frame->linesize[0] = stride[0];
+        if (params.force_yuv) {
+            convert_pixels_to_yuv(pixels, formatted_pixels, stride);
+        } else {
+            encoder_frame->data[0] = (uint8_t*) formatted_pixels;
+            encoder_frame->linesize[0] = stride[0];
+        }
 
         if (av_hwframe_transfer_data(hw_frame, encoder_frame, 0))
         {
@@ -387,8 +496,8 @@ void FrameWriter::add_frame(const uint8_t* pixels, int64_t usec, bool y_invert)
         encoder_frame->buf[0] = NULL;
     } else
     {
-        sws_scale(swsCtx, &formatted_pixels, stride, 0, params.height,
-            encoder_frame->data, encoder_frame->linesize);
+        convert_pixels_to_yuv(pixels, formatted_pixels, stride);
+
         /* Force ffmpeg to create a copy of the frame, if the codec needs it */
         saved_buf0 = encoder_frame->buf[0];
         encoder_frame->buf[0] = NULL;
@@ -402,15 +511,11 @@ void FrameWriter::add_frame(const uint8_t* pixels, int64_t usec, bool y_invert)
     pkt.data = NULL;
     pkt.size = 0;
 
-    int got_output;
-    avcodec_encode_video2(videoCodecCtx, &pkt, *output_frame, &got_output);
+    encode(videoCodecCtx, *output_frame, &pkt);
 
     /* Restore frame buffer, so that it can be properly freed in the end */
     if (saved_buf0)
         encoder_frame->buf[0] = saved_buf0;
-
-    if (got_output)
-      finish_frame(pkt, true);
 }
 
 #define SRC_RATE 1e6
@@ -437,10 +542,7 @@ void FrameWriter::send_audio_pkt(AVFrame *frame)
     pkt.data = NULL;
     pkt.size = 0;
 
-    int got_output;
-    avcodec_encode_audio2(audioCodecCtx, &pkt, frame, &got_output);
-    if (got_output)
-      finish_frame(pkt, false);
+    encode(audioCodecCtx, frame, &pkt);
 }
 
 size_t FrameWriter::get_audio_buffer_size()
@@ -475,11 +577,11 @@ void FrameWriter::add_audio(const void* buffer)
     av_frame_free(&outputf);
 }
 
-void FrameWriter::finish_frame(AVPacket& pkt, bool is_video)
+void FrameWriter::finish_frame(AVCodecContext *enc_ctx, AVPacket& pkt)
 {
     static std::mutex fmt_mutex, pending_mutex;
 
-    if (is_video)
+    if (enc_ctx == videoCodecCtx)
     {
         av_packet_rescale_ts(&pkt, (AVRational){ 1, 1000000 }, videoStream->time_base);
         pkt.stream_index = videoStream->index;
@@ -512,25 +614,18 @@ FrameWriter::~FrameWriter()
     AVPacket pkt;
     av_init_packet(&pkt);
 
-    for (int got_output = 1; got_output;)
-    {
-        avcodec_encode_video2(videoCodecCtx, &pkt, NULL, &got_output);
-        if (got_output)
-            finish_frame(pkt, true);
-    }
+    encode(videoCodecCtx, NULL, &pkt);
 
-    for (int got_output = 1; got_output && params.enable_audio;)
+    if (params.enable_audio)
     {
-        avcodec_encode_audio2(audioCodecCtx, &pkt, NULL, &got_output);
-        if (got_output)
-            finish_frame(pkt, false);
+        encode(audioCodecCtx, NULL, &pkt);
     }
 
     // Writing the end of the file.
     av_write_trailer(fmtCtx);
 
     // Closing the file.
-    if (!(outputFmt->flags & AVFMT_NOFILE))
+    if (outputFmt && (!(outputFmt->flags & AVFMT_NOFILE)))
         avio_closep(&fmtCtx->pb);
 
     avcodec_close(videoStream->codec);
