@@ -111,7 +111,7 @@ struct wf_buffer
     bool y_invert;
 
     timespec presented;
-    uint32_t base_usec;
+    uint64_t base_usec;
 
     std::atomic<bool> released{true}; // if the buffer can be used to store new pending frames
     std::atomic<bool> available{false}; // if the buffer can be used to feed the encoder
@@ -609,7 +609,7 @@ int main(int argc, char *argv[])
         { "help",            no_argument,       NULL, 'h' },
         { "force-yuv",       no_argument,       NULL, 't' },
         { "opencl",          optional_argument, NULL, 'e' },
-        { "bframes",         optional_argument, NULL, 'b' },
+        { "bframes",         required_argument, NULL, 'b' },
         { 0,                 0,                 NULL,  0  }
     };
 
@@ -649,7 +649,7 @@ int main(int argc, char *argv[])
                 break;
 
             case 'b':
-                params.bframes = optarg ? atoi(optarg) : -1;
+                params.bframes = atoi(optarg);
                 break;
 
             case 'l':
